module Main where 

-- import the Monad and MonadPlus classes
import Monad   

-- import various character utilities 
import Char (isDigit,ord,isLower,isUpper,isAlpha,isSpace)

------------------------------------------------------------------------------
---                                 Parsers                                ---
------------------------------------------------------------------------------

-- The monad of parsers: -------------------------------------------------

newtype Parser a = Parser (String -> [(a,String)])

instance Monad Parser where 
   return a      = Parser (\cs -> [(a,cs)])
   p >>= f       = Parser (\cs -> concat [parse (f a) cs' |
                                     (a,cs') <- parse p cs])

instance MonadPlus Parser where
   mzero         = Parser (\cs -> [])
   p `mplus` q   = Parser (\cs -> parse p cs ++ parse q cs)


-- Primitive parsers: -----------------------------------------

parse           :: Parser a -> String -> [(a,String)]
parse (Parser p) = p

item            :: Parser Char
item             = Parser (\cs -> case cs of
                                     ""     -> []
                                     (c:cs) -> [(c,cs)])

sat             :: (Char -> Bool) -> Parser Char
sat p            = do {c <- item; if p c then return c else mzero}


-- Efficiency improving combinators: ---------------------------------

(+++)           :: Parser a -> Parser a -> Parser a
p +++ q          = Parser (\cs -> case parse (p `mplus` q) cs of
                                     []     -> []
                                     (x:xs) -> [x])

-- Recursion combinators: --------------------------------------------

string          :: String -> Parser String
string ""        = return ""
string (c:cs)    = do {char c; string cs; return (c:cs)}

many            :: Parser a -> Parser [a]
many p           = many1 p +++ return []

many1           :: Parser a -> Parser [a]
many1 p          = do {a <- p; as <- many p; return (a:as)}

sepby           :: Parser a -> Parser b -> Parser [a]
p `sepby` sep    = (p `sepby1` sep) +++ return []

sepby1          :: Parser a -> Parser b -> Parser [a]
p `sepby1` sep   = do {a <- p; as <- many (do {sep; p}); return (a:as)}

chainl          :: Parser a -> Parser (a -> a -> a) -> a -> Parser a
chainl p op a    = (p `chainl1` op) +++ return a

chainl1         :: Parser a -> Parser (a -> a -> a) -> Parser a
p `chainl1` op   = do {a <- p; rest a}
                   where
                      rest a = do {f <- op; b <- p; rest (f a b)}
                               +++ return a

------------------------------------------------------------------------------
---                          Basic XML Definitions                         ---
------------------------------------------------------------------------------

data XML =                                             
         PCDATA  String                                                     
       | Element ElementName [Attribute] [XML]
     deriving (Eq,Show)

type ElementName = String                                                    
type Attribute = (String,String)


------------------------------------------------------------------------------
---                           Homework Solutions                           ---
------------------------------------------------------------------------------

-- Arithmetic expression parsers ---------------------------------------------

char            :: Char -> Parser Char
char c           = sat (c ==)

numchar         :: Parser Char
numchar          = sat isDigit

digit           :: Parser Int
digit            = do {c <- numchar; return (ord c - ord '0')}

natural         :: Parser Int
natural          = digit `chainl1` return (\m n -> 10*m + n)


expr  :: Parser Int
addop :: Parser (Int -> Int -> Int)
mulop :: Parser (Int -> Int -> Int)

expr   = term   `chainl1` addop
term   = factor `chainl1` mulop
factor = natural +++ do {char '('; n <- expr; char ')'; return n}

addop  = do {char '+'; return (+)} +++ do {char '-'; return (-)}
mulop  = do {char '*'; return (*)} +++ do {char '/'; return (div)}


-- XML parsers----------------------------------------------------------------

xmlelt :: Parser XML
xmlelt = return (PCDATA "Write me!")


------------------------------------------------------------------------------
---                               Testing                                  ---
------------------------------------------------------------------------------

-- Testing function (do not modify)
test name parser a b =
  do putStr (take 20 (name ++ repeat ' ')); 
     case (parse parser a) of
       [] -> putStr " failed with no results\n" 
       (x,""):_  -> 
           if x==b 
             then putStr " succeeded\n" 
             else do putStr " failed:  expected "; putStr (show a)
                     putStr " but got "; print b
       (x,s):_ -> do putStr " failed:  parsed "; putStr (show x);
                     putStr " but '"; putStr s; putStr "' was left over\n"

-- Do not modify this definition:
officialTests = do 
  -- Regression test for the arithmetic expression parser
  test "basic expr test 1"
    expr "2+3*5"
    17
  test "basic expr test 2"
    expr "(2+3)*5"
    25
  -- Tests for solutions to expression part of assignment
  test "negative 1"
    expr "3*(-2)"
    (-6)
  test "negative 2"
    expr "3*-2"
    (-6)
  test "whitespace 1"
    expr "2 + 3   *5"
    17
  test "whitespace 2"
    expr " 2 + 3   *5  "
    17
  test "exponentiation"
    expr "3*2^2"
    12
  -- Tests for XML parsing
  test "xml 1" 
    xmlelt ("<A></A>")
    (Element "A" [] [])
  test "xml 2" 
    xmlelt ("<A><B>hello</B>there</A>")
    (Element "A" [] [Element "B" [] [PCDATA "hello"], PCDATA "there"])
  test "xml 3" 
    xmlelt ("<A x=\"5\"></A>")
    (Element "A" [("x","5")] [])
  test "xml 4" 
    xmlelt (   "<A>data1<B a1=\"v1\" a2=\"v2\">moredata</B>data2data3"
            ++ "<C a3=\"v3\"></C>data4</A>")
    (Element "A" [] 
      [PCDATA "data1",
       Element "B" [("a1","v1"),("a2","v2")] [PCDATA "moredata"], 
       PCDATA "data2data3",
       Element "C" [("a3","v3")] [],
       PCDATA "data4"])

-- Feel free to modify this during development, but make sure to put
-- it back exactly like this before you hand in your assignment.
main = officialTests
